/* Written by Yong Li                                                  */
/* Copyright (c) 2016                  	                               */
/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License    */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.*/

package cn.ac.ios.roll.inclusion;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import automata.FAState;
import automata.FiniteAutomaton;
import cn.ac.ios.automata.Automaton;
import cn.ac.ios.automata.words.Alphabet;
import cn.ac.ios.automata.words.WordManager;
import cn.ac.ios.automata.parser.FormatType;
import cn.ac.ios.automata.parser.ba.BAParser;
import cn.ac.ios.automata.parser.hoa.HanoiParser;
import cn.ac.ios.learner.Learner;
import cn.ac.ios.learner.buechi.ldollar.LearnerOmegaBuechi;
import cn.ac.ios.learner.dfa.table.LearnerDFATable;
import cn.ac.ios.learner.dfa.tree.LearnerDFATree;
import cn.ac.ios.learner.fdfa.LearnerFDFA;
import cn.ac.ios.learner.fdfa.table.LearnerFDFATablePeriodic;
import cn.ac.ios.learner.fdfa.table.LearnerFDFATableRecurrent;
import cn.ac.ios.learner.fdfa.table.LearnerFDFATableSyntactic;
import cn.ac.ios.learner.fdfa.tree.LearnerFDFATreePeriodic;
import cn.ac.ios.learner.fdfa.tree.LearnerFDFATreeRecurrent;
import cn.ac.ios.learner.fdfa.tree.LearnerFDFATreeSyntactic;
import cn.ac.ios.learner.lstar.LearnerLStar;
import cn.ac.ios.log.Log;
import cn.ac.ios.options.Options;
import cn.ac.ios.oracle.BuechiRunner;
import cn.ac.ios.oracle.EquivalenceOracleImpl;
import cn.ac.ios.oracle.MembershipOracleImpl;
import cn.ac.ios.oracle.Minimizer;
import cn.ac.ios.oracle.teacher.Teacher;
import cn.ac.ios.oracle.teacher.DFATeacherDK;
import cn.ac.ios.oracle.teacher.BuechiTeacherRabit;
import cn.ac.ios.oracle.teacher.BuechiTeacherSample;
import cn.ac.ios.query.EquivalenceOracle;
import cn.ac.ios.query.MembershipOracle;
import cn.ac.ios.query.Query;
import cn.ac.ios.roll.Statistics;
import cn.ac.ios.roll.sampler.FASampler;
import cn.ac.ios.roll.test.AutomatonGenerator;
import cn.ac.ios.roll.translator.Translator;
import cn.ac.ios.roll.translator.TranslatorBuechi;
import cn.ac.ios.roll.translator.TranslatorFDFAUnder;
import cn.ac.ios.roll.translator.TranslatorFDFAOver;
import cn.ac.ios.roll.translator.TranslatorSimple;
import cn.ac.ios.util.AutomatonPrinter;
import cn.ac.ios.util.BuechiBuilder;
import cn.ac.ios.util.InputHelper;
import cn.ac.ios.util.Timer;
import cn.ac.ios.util.UtilAutomaton;
import cn.ac.ios.value.ValueManager;
import dk.brics.automaton.State;
import cn.ac.ios.value.Type;
import cn.ac.ios.value.Value;
import oracle.EqResult;
import oracle.IntersectionCheck;
import oracle.RabitOracle;


/**
Algorithm to check the inclusion for two given Buechi automata A and B
i.e., L(A) <= L(B)
 */

/**
 * @author Yong Li
 *  */
public class BuechiInclusion 
{

	

	public static BAParser parser = null;
    public static void main(String[] args)
    {
		Options.log = new Log(System.out);
		FiniteAutomaton A = null;
		FiniteAutomaton B = null;
		int numST = 0;
		int numTests = 0;
    	// parse input arguments
    	for(int i = 0; i < args.length; i ++) {
    		if(args[i].compareTo("-h")==0) {
    			helper();
    		}
    	}
    	
    	if(args.length < 2) {
    		helper();
    	}

    	// for learning;
    	for(int i = 0; i < args.length; i ++) {
			
			if(args[i].compareTo("-v")==0){
				Options.verbose=true;
				continue;
			}
			
			if(args[i].compareTo("-table") == 0) {
    			Options.learnerTree = false;
    			Options.learnerTable = true;
    			continue;
			}
			if(args[i].compareTo("-tree") == 0) {
    			Options.learnerTree = true;
    			Options.learnerTable = false;
    			continue;
			}
			if(args[i].compareTo("-lstar") == 0) {
				Options.learnerLStar = true;
				Options.learnerMultiplicity = false;
				Options.learnerDFA = false;
				Options.learnerNFA = false;
				Options.learnerLDollar = false;
				Options.learnerPeriodic = false;
				Options.learnerRecurrent = false;
				Options.learnerSyntactic = false;
				Options.lazyEqCheck = false;
				continue;
			}
			if(args[i].compareTo("-dfa") == 0) {
				Options.learnerDFA = true;
				Options.learnerLStar = false;
				Options.learnerMultiplicity = false;
				Options.learnerNFA = false;
				Options.learnerLDollar = false;
				Options.learnerPeriodic = false;
				Options.learnerRecurrent = false;
				Options.learnerSyntactic = false;
				Options.lazyEqCheck = false;
				continue;
			}
			if(args[i].compareTo("-ldollar") == 0) {
				Options.learnerLStar = false;
				Options.learnerMultiplicity = false;
				Options.learnerNFA = false;
				Options.learnerDFA = false;
				Options.learnerLDollar = true;
				Options.learnerPeriodic = false;
				Options.learnerRecurrent = false;
				Options.learnerSyntactic = false;
				continue;
			}
			if(args[i].compareTo("-syntactic") == 0) {
				Options.learnerLStar = false;
				Options.learnerMultiplicity = false;
				Options.learnerNFA = false;
				Options.learnerDFA = false;
				Options.learnerLDollar = false;
				Options.learnerPeriodic = false;
				Options.learnerRecurrent = false;
				Options.learnerSyntactic = true;
				continue;
			}
			if(args[i].compareTo("-recurrent") == 0) {
				Options.learnerLStar = false;
				Options.learnerMultiplicity = false;
				Options.learnerDFA = false;
				Options.learnerNFA = false;
				Options.learnerLDollar = false;
				Options.learnerPeriodic = false;
				Options.learnerRecurrent = true;
				Options.learnerSyntactic = false;
				continue;
			}
			if(args[i].compareTo("-periodic") == 0) {
				Options.learnerLStar = false;
				Options.learnerMultiplicity = false;
				Options.learnerDFA = false;
				Options.learnerNFA = false;
				Options.learnerLDollar = false;
				Options.learnerPeriodic = true;
				Options.learnerRecurrent = false;
				Options.learnerSyntactic = false;
				continue;
			}
			
			if(args[i].compareTo("-under") == 0) {
				Options.buechiBuildUnder = true;
				continue;
			}
			if(args[i].compareTo("-sim") == 0) {
				Options.simulationNFA = true;
				continue;
			}
			
			if(args[i].compareTo("-min") == 0) {
				Options.minimizationBuechi = true;
				continue;
			}
			
			if(args[i].compareTo("-over") == 0) {
				Options.buechiBuildUnder = false;
				Options.outLDBA = false;
				continue;
			}

			
			if(args[i].compareTo("-lazyeq") == 0) {
				Options.lazyEqCheck = true;
				continue;
			}

			
			if(args[i].compareTo("-dot") == 0) {
				Options.dot = true;
				continue;
			}
			
			if(args[i].endsWith(".ba")) {
				Options.formatType = FormatType.BA;
				
				if(A == null) {
					parser = new BAParser(args[i]);
					A = parser.parse();
				}else {
					parser = new BAParser(args[i], parser); // share the previous alphabet
					B = parser.parse();
				}
				continue;
			}
			
			
			if(args[i].compareTo("-test") == 0) {
				Options.modeAutomatic = false;
				Options.modeInteractive = false;
				Options.modeTest = true;
				Options.numTests = Integer.parseInt(args[i+1]);
				Options.numStatesForTest = Integer.parseInt(args[i + 2]);
				i += 2;
				continue;
			}
    	}

    	
    	Statistics.reset();

		if(Options.modeTest) { // do 20 tests?
			int num = 0;
			for(int n = 0; n < numTests; n ++) {
				FiniteAutomaton aut = AutomatonGenerator.getRandomAutomaton(numST, 3);
				
				try{
					System.out.println("target: " );
					AutomatonPrinter.print(aut, System.out);
					System.out.println("complementation: " );
					AutomatonPrinter.print(Statistics.hypothesis, System.out);
					if(validate(Statistics.hypothesis, aut)) {
						++ num;
					}
				}catch (Exception e)
				{
					e.printStackTrace();
					aut.saveAutomaton("bug.ba");
					System.exit(-1);
				}
				aut.saveAutomaton("example.ba");
				System.out.println("Done for case " + (n + 1));
			}
			System.out.println(num + " out of " + numTests + " cases are correct");
		}else {
			decideInclusion(A, B, parser);
		}


    	
    }
		
	private static void decideInclusion(FiniteAutomaton A, FiniteAutomaton B, BAParser parser) {
		// check automata inclusion
		ValueManager contextValue = new ValueManager();
		WordManager manager = new WordManager(contextValue);
		
		Set<String> letters = new HashSet<>();
		for(String letter : A.alphabet) {
			letters.add(letter);
		}
		
		for(String letter : B.alphabet) {
			letters.add(letter);
		}
		
		// initialize alphabet
		Type typeString = contextValue.newTypeLetter(String.class);
		Value valueLetter = typeString.newValue();
		Alphabet alphabet = manager.getAlphabet();
		if(Options.learnerLDollar) {
			valueLetter.set(WordManager.getStringDollar());
			alphabet.add(valueLetter.clone());
		}
		for(String letter : letters) {
			valueLetter.set(letter);
			alphabet.add(valueLetter.clone());
		}
		System.out.println("alphabet size: " + alphabet.size());
		alphabet.setImmutable();
		valueLetter.set("");
	    manager.setLetterSplitter(valueLetter);
	    
//	    System.out.println("A: ");
//	    AutomatonPrinter.print(A, System.out);
	    
//	    System.out.println("B: ");
//	    AutomatonPrinter.print(B, System.out);
	    
		InclusionEQOracle eqOracle = new InclusionEQOracle(manager, A, B, parser);
		LearnerFDFA learner = pickOneLearner(manager, new MQOracle(manager, B));
		// learning loop
		Timer timerTotal = new Timer();
		timerTotal.start();
		System.out.println("Start learning...");
		Timer timer = new Timer();
		timer.start();
		learner.startLearning();
		timer.stop();
		Statistics.timeLearner += timer.getTimeElapsed();
		boolean result = false;
		while(! result ) {
			if(Options.verbose) System.out.println("learner output: " + learner.toString());
			
			Query<EqResult> query = eqOracle.answerEquivalenceQuery(learner);
			// get out of the loop
			if(query.getQueryAnswer().isEqual == true) {
				Statistics.setLearnerFDFA(learner);
//				if(outputFile != null) Statistics.hypothesis.saveAutomaton(outputFile);
				break;
			}
			
			// lazy equivalence check is implemented here
			Translator translator = new TranslatorFDFAUnder(learner);
			translator.setQuery(query);
			while(translator.canRefine()) {
				Query<Boolean> ceQuery = translator.translate();
				timer.start();
				learner.refineHypothesis(ceQuery);
				timer.stop();
				Statistics.timeLearner += timer.getTimeElapsed();
				if(Options.verbose) System.out.println("learner output: " + learner.toString());
				// if do not set lazy eq check or it is learnerBuechi
				if(! Options.lazyEqCheck) break;
			}

		}
		timerTotal.stop();
		System.out.println("Learning completed...");
		eqOracle.print();
		System.out.println(timerTotal.getTimeElapsed()+ " ms elapsed...");
	}


	private static LearnerFDFA pickOneLearner(WordManager contextWord,
    		MembershipOracle<Boolean> membershipOracle) {
    	
    	if(Options.learnerTable) {
    		if(Options.learnerRecurrent) return new LearnerFDFATableRecurrent(contextWord, membershipOracle);
    		if(Options.learnerSyntactic) return new LearnerFDFATableSyntactic(contextWord, membershipOracle);
    		return new LearnerFDFATablePeriodic(contextWord, membershipOracle);

    	}else {
    		
    		if(Options.learnerRecurrent) return new LearnerFDFATreeRecurrent(contextWord, membershipOracle);
    		if(Options.learnerSyntactic) return new LearnerFDFATreeSyntactic(contextWord, membershipOracle);
    		return new LearnerFDFATreePeriodic(contextWord, membershipOracle);
    	}
    }
    
    public static void helper() {
		System.out.println(
				"BUNE v1 (BUechi inclusioN chEck base on learning)");
		System.out.println(
				"Usage: java -jar bune.jar [aut1.ba, aut2.ba] [options]");

		System.out.println("             or: java -jar bune.jar aut.ba -table -periodic");
		System.out.println("             or: java -jar bune.jar aut.hoa -table -periodic");
		System.out.println("             or: java -jar bune.jar -t 3 3 -table -periodic");
		System.out.println("\nOptions:");
		System.out.println("-h: Show this page.");
		System.out.println("-v: Verbose mode.");
		System.out.println(
				"-t k n: Test BUNE with k randomly generated Buechi automata of n states.");
		System.out.println(
				"-tree: Use tree-based data structure in the algorithm (Default).");
		System.out.println(
				"-table: Use table-based data structure in the algorithm.");
		System.out.println(
				"-periodic: Use periodic FDFA to learn Omega regular language.");
		System.out.println(
				"-syntactic: Use syntactic FDFA to learn Omega regular language.");
		System.out.println(
				"-recurrent: Use recurrent FDFA to learn Omega regular language.");
//		System.out.println(
//				"-samfdfa: Sampling the FDFA before equivalence check. (Not yet supported)");
		System.out.println(
				"-sim: Use Similation provided by RABIT to reduce automata before intersection of DFAs.");
		System.out.println(
				"-min: Use Minimization provided by RABIT to reduce Buechi automata before equivalence check.");
		System.out.println(
				"-lazyeq: Equivalence check as the last resort. ");
		System.exit(0);
    }
    
 
    private static boolean validate(FiniteAutomaton a, FiniteAutomaton b) {
    	FiniteAutomaton c = new FiniteAutomaton();
    	
    	FAState q = c.createState();
    	for(String letter : a.alphabet) {
    		c.addTransition(q, q, letter);
    	}
    	
    	c.setInitialState(q);
    	c.F.add(q);
    	
    	dk.brics.automaton.Automaton dkA = UtilAutomaton.convertToDkAutomaton(a);
    	dk.brics.automaton.Automaton dkB = UtilAutomaton.convertToDkAutomaton(b);
    	
    	dkA = dkA.union(dkB);
    	
        FiniteAutomaton ck = UtilAutomaton.convertToRabitAutomaton(dkA);
		PrintStream outOld = System.out;
		try {
			PrintStream out = new PrintStream(new FileOutputStream("/dev/null"));
			System.setOut(out);
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
        RabitOracle oracle = new RabitOracle(c);
        boolean result = oracle.isEqualBuechi(ck).isEqual;
        System.setOut(outOld);
        System.out.println("Complementation result is " + result);
        
        return result;
    }
  
}
