/* Written by Yong Li, Depeng Liu                                       */
/* Copyright (c) 2016                  	                               */
/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License    */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.*/

package cn.ac.ios.dk;

import automata.FiniteAutomaton;
import cn.ac.ios.oracle.Minimizer;
import cn.ac.ios.util.AutomatonPrinter;
import cn.ac.ios.util.LDBABuilder;
import cn.ac.ios.util.UtilAutomaton;
import dk.brics.automaton.Automaton;
import dk.brics.automaton.State;
import dk.brics.automaton.Transition;

public class TestDK {
	

	
	private static Automaton constructM() {
		Automaton m = new Automaton();
		
		State init = new State();
        State acc = new State();
        
		init.addTransition(new Transition('a', acc));
		init.addTransition(new Transition('b', init));
		acc.addTransition(new Transition('a', init));
		acc.addTransition(new Transition('b', acc));
		

		
		m.setInitialState(init);
		
		return m;
	}
	
	private static Automaton constructL() {
		Automaton m = new Automaton();
		
		State init = new State();
        State acc = new State();
        
		init.addTransition(new Transition('b', init));
		init.addTransition(new Transition('a', acc));
		acc.addTransition(new Transition('a', acc));
		acc.addTransition(new Transition('b', init));
		

		
		m.setInitialState(init);
		acc.setAccept(true);
		return m;
	}
	
	private static Automaton constructA() {
		Automaton m = new Automaton();
		
		State init = new State();
        State acc = new State();
        
		init.addTransition(new Transition('b', acc));
		init.addTransition(new Transition('a', acc));
		acc.addTransition(new Transition('a', acc));
		acc.addTransition(new Transition('b', init));
		

		
		m.setInitialState(init);
		acc.setAccept(true);
		return m;
	}
	
	public static void main(String []args) {
		Automaton m = constructM();
		System.out.println("M : " + m.toDot());
		Automaton l = constructL();
		System.out.println("L : " + l.toDot());
		Automaton a = constructA();
		System.out.println("A : " + a.toDot());
		
		Automaton ml = m.clone();
		ml.getInitialState().setAccept(true);
		System.out.println("ML-P : " + ml.toDot());
		ml = ml.intersection(l.clone());
		System.out.println("ML-A : " + ml.toDot());
		
		Automaton ma = m.clone();
		ma.getInitialState().step('a').setAccept(true);		
		System.out.println("MA-P : " + ma.toDot());
		ma = ma.intersection(a.clone());
		System.out.println("MA-A : " + ma.toDot());
		
		Automaton ml2 = l.clone();
		ml2.setInitialState(ml2.getInitialState().step('a'));
		System.out.println("A-V : " + ml2.toDot());
		ml2 = ml2.intersection(ml);
		System.out.println("MA-V : " + ml2.toDot());
		String ce = ml2.minus(ml).getShortestExample(true);
		System.out.println("MA-V- : " + ce);
		
		Automaton ma2 = a.clone();
		ma2.setInitialState(ma2.getInitialState().step('a'));
		System.out.println("A-V : " + ma2.toDot());
		ma2 = ma2.intersection(ma);
		System.out.println("MA-V : " + ma2.toDot());
		ce = ma2.minus(ma).getShortestExample(true);
		System.out.println("MA-V- : " + ce);
		
	}

}
